@@ -12,6 +12,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <stdbool.h>
                      
 #include <netlink/genl/genl.h>
 #include <netlink/genl/family.h>
@@ -21,7 +22,6 @@
 
 #include "nl80211.h"
 #include "iw.h"
-#include "version.h"
 
 #ifndef CONFIG_LIBNL20
 /* libnl 2.0 compatibility code */
@@ -47,7 +47,7 @@ static inline int __genl_ctrl_alloc_cache(struct nl_sock *h, struct nl_cache **c
 #define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
 #endif /* CONFIG_LIBNL20 */
 
-static int debug = 0;
+int iw_debug = 0;
 
 static int nl80211_init(struct nl80211_state *state)
 {
@@ -94,49 +94,106 @@ static void nl80211_cleanup(struct nl80211_state *state)
 	nl_socket_free(state->nl_sock);
 }
 
-__COMMAND(NULL, NULL, NULL, 0, 0, 0, CIB_NONE, NULL);
-__COMMAND(NULL, NULL, NULL, 1, 0, 0, CIB_NONE, NULL);
+__COMMAND(NULL, NULL, "", NULL, 0, 0, 0, CIB_NONE, NULL, NULL);
+__COMMAND(NULL, NULL, "", NULL, 1, 0, 0, CIB_NONE, NULL, NULL);
 
 static int cmd_size;
 
-static void usage(const char *argv0)
+static void __usage_cmd(struct cmd *cmd, char *indent, bool full)
+{
+	const char *start, *lend, *end;
+
+	fprintf(stderr, "%s", indent);
+
+	switch (cmd->idby) {
+	case CIB_NONE:
+		break;
+	case CIB_PHY:
+		fprintf(stderr, "phy <phyname> ");
+		break;
+	case CIB_NETDEV:
+		fprintf(stderr, "dev <devname> ");
+		break;
+	}
+	if (cmd->section)
+		fprintf(stderr, "%s ", cmd->section);
+	fprintf(stderr, "%s", cmd->name);
+	if (cmd->args)
+		fprintf(stderr, " %s", cmd->args);
+	fprintf(stderr, "\n");
+
+	if (!full || !cmd->help)
+		return;
+
+	/* hack */
+	if (strlen(indent))
+		indent = "\t\t";
+	else
+		fprintf(stderr, "\n");
+
+	/* print line by line */
+	start = cmd->help;
+	end = strchr(start, '\0');
+	do {
+		lend = strchr(start, '\n');
+		if (!lend)
+			lend = end;
+		fprintf(stderr, "%s", indent);
+		fprintf(stderr, "%.*s\n", (int)(lend - start), start);
+		start = lend + 1;
+	} while (end != lend);
+
+	fprintf(stderr, "\n");
+}
+
+static void usage_options(void)
+{
+	fprintf(stderr, "Options:\n");
+	fprintf(stderr, "\t--debug\t\tenable netlink debugging\n");
+}
+
+static const char *argv0;
+
+static void usage(bool full)
 {
 	struct cmd *cmd;
 
 	fprintf(stderr, "Usage:\t%s [options] command\n", argv0);
-	fprintf(stderr, "Options:\n");
-	fprintf(stderr, "\t--debug\t\tenable netlink debugging\n");
-	fprintf(stderr, "\t--version\tshow version\n");
+	usage_options();
+	fprintf(stderr, "\t--version\tshow version (%s)\n", iw_version);
 	fprintf(stderr, "Commands:\n");
-	fprintf(stderr, "\thelp\n");
-	fprintf(stderr, "\tevent\n");
 	for (cmd = &__start___cmd; cmd < &__stop___cmd;
 	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
 		if (!cmd->handler || cmd->hidden)
 			continue;
-		switch (cmd->idby) {
-		case CIB_NONE:
-			fprintf(stderr, "\t");
-			break;
-		case CIB_PHY:
-			fprintf(stderr, "\tphy <phyname> ");
-			break;
-		case CIB_NETDEV:
-			fprintf(stderr, "\tdev <devname> ");
-			break;
-		}
-		if (cmd->section)
-			fprintf(stderr, "%s ", cmd->section);
-		fprintf(stderr, "%s", cmd->name);
-		if (cmd->args)
-			fprintf(stderr, " %s", cmd->args);
-		fprintf(stderr, "\n");
+		__usage_cmd(cmd, "\t", full);
 	}
+	fprintf(stderr, "\nYou can omit the 'phy' or 'dev' if "
+			"the identification is unique,\n"
+			"e.g. \"iw wlan0 info\" or \"iw phy0 info\". "
+			"(Don't when scripting.)\n\n");
+}
+
+static int print_help(struct nl80211_state *state,
+		      struct nl_cb *cb,
+		      struct nl_msg *msg,
+		      int argc, char **argv)
+{
+	exit(3);
+}
+TOPLEVEL(help, NULL, 0, 0, CIB_NONE, print_help,
+	 "Print usage for each command.");
+
+static void usage_cmd(struct cmd *cmd)
+{
+	fprintf(stderr, "Usage:\t%s [options] ", argv0);
+	__usage_cmd(cmd, "", true);
+	usage_options();
 }
 
 static void version(void)
 {
-	printf("iw version " IW_VERSION "\n");
+	printf("iw version %s\n", iw_version);
 }
 
 static int phy_lookup(char *name)
@@ -178,8 +235,8 @@ static int ack_handler(struct nl_msg *msg, void *arg)
 	return NL_STOP;
 }
 
-int handle_cmd(struct nl80211_state *state, enum id_input idby,
-	       int argc, char **argv)
+static int __handle_cmd(struct nl80211_state *state, enum id_input idby,
+			int argc, char **argv, struct cmd **cmdout)
 {
 	struct cmd *cmd, *match = NULL;
 	struct nl_cb *cb;
@@ -265,6 +322,9 @@ int handle_cmd(struct nl80211_state *state, enum id_input idby,
 	if (!cmd)
 		return 1;
 
+	if (cmdout)
+		*cmdout = cmd;
+
 	if (!cmd->cmd) {
 		argc = o_argc;
 		argv = o_argv;
@@ -277,7 +337,7 @@ int handle_cmd(struct nl80211_state *state, enum id_input idby,
 		return 2;
 	}
 
-	cb = nl_cb_alloc(debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
+	cb = nl_cb_alloc(iw_debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
 	if (!cb) {
 		fprintf(stderr, "failed to allocate netlink callbacks\n");
 		err = 2;
@@ -324,175 +384,27 @@ int handle_cmd(struct nl80211_state *state, enum id_input idby,
 	return 2;
 }
 
-static int no_seq_check(struct nl_msg *msg, void *arg)
+int handle_cmd(struct nl80211_state *state, enum id_input idby,
+	       int argc, char **argv)
 {
-	return NL_OK;
-}
-
-static int print_event(struct nl_msg *msg, void *arg)
-{
-	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
-	struct nlattr *tb[NL80211_ATTR_MAX + 1];
-	char ifname[100];
-	__u8 reg_type;
-
-	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
-		  genlmsg_attrlen(gnlh, 0), NULL);
-                          
-	switch (gnlh->cmd) {
-	case NL80211_CMD_NEW_WIPHY:
-		printf("wiphy rename: phy #%d to %s\n",
-		       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
-		       nla_get_string(tb[NL80211_ATTR_WIPHY_NAME]));
-		break;
-	case NL80211_CMD_NEW_SCAN_RESULTS:
-		if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
-		printf("scan finished on %s (phy #%d)\n",
-		       ifname, nla_get_u32(tb[NL80211_ATTR_WIPHY]));
-		break;
-	case NL80211_CMD_SCAN_ABORTED:
-		if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
-		printf("scan aborted on %s (phy #%d)\n",
-		       ifname, nla_get_u32(tb[NL80211_ATTR_WIPHY]));
-		break;
-	case NL80211_CMD_REG_CHANGE:
-
-		printf("regulatory domain change: ");
-
-		reg_type = nla_get_u8(tb[NL80211_ATTR_REG_TYPE]);
-
-		switch (reg_type) {
-		case NL80211_REGDOM_TYPE_COUNTRY:
-			printf("set to %s by %s request",
-			       nla_get_string(tb[NL80211_ATTR_REG_ALPHA2]),
-			       reg_initiator_to_string(nla_get_u8(tb[NL80211_ATTR_REG_INITIATOR])));
-			if (tb[NL80211_ATTR_WIPHY])
-				printf(" on phy%d", nla_get_u32(tb[NL80211_ATTR_WIPHY]));
-			break;
-		case NL80211_REGDOM_TYPE_WORLD:
-			printf("set to world roaming by %s request",
-			       reg_initiator_to_string(nla_get_u8(tb[NL80211_ATTR_REG_INITIATOR])));
-			break;
-		case NL80211_REGDOM_TYPE_CUSTOM_WORLD:
-			printf("custom world roaming rules in place on phy%d by %s request",
-			       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
-			       reg_initiator_to_string(nla_get_u32(tb[NL80211_ATTR_REG_INITIATOR])));
-			break;
-		case NL80211_REGDOM_TYPE_INTERSECTION:
-			printf("intersection used due to a request made by %s",
-			       reg_initiator_to_string(nla_get_u32(tb[NL80211_ATTR_REG_INITIATOR])));
-			if (tb[NL80211_ATTR_WIPHY])
-				printf(" on phy%d", nla_get_u32(tb[NL80211_ATTR_WIPHY]));
-			break;
-		default:
-			printf("unknown source (upgrade this utility)");
-			break;
-		}
-
-		printf("\n");
-		break;
-	default:
-		printf("unknown event: %d\n", gnlh->cmd);
-		break;
-	}
-
-	return NL_SKIP;
-}
-
-struct wait_event {
-	int n_cmds;
-	const __u32 *cmds;
-	__u32 cmd;
-};
-
-static int wait_event(struct nl_msg *msg, void *arg)
-{
-	struct wait_event *wait = arg;
-	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
-	int i;
-
-	for (i = 0; i < wait->n_cmds; i++) {
-		if (gnlh->cmd == wait->cmds[i]) {
-			wait->cmd = gnlh->cmd;
-		}
-	}
-
-	return NL_SKIP;
-}
-
-__u32 listen_events(struct nl80211_state *state,
-		    const int n_waits, const __u32 *waits)
-{
-	int mcid, ret;
-	struct nl_cb *cb = nl_cb_alloc(debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
-	struct wait_event wait_ev;
-
-	if (!cb) {
-		fprintf(stderr, "failed to allocate netlink callbacks\n");
-		return -ENOMEM;
-	}
-
-	/* Configuration multicast group */
-	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "config");
-	if (mcid < 0)
-		return mcid;
-
-	ret = nl_socket_add_membership(state->nl_sock, mcid);
-	if (ret)
-		return ret;
-
-	/* Scan multicast group */
-	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "scan");
-	if (mcid >= 0) {
-		ret = nl_socket_add_membership(state->nl_sock, mcid);
-		if (ret)
-			return ret;
-	}
-
-	/* Regulatory multicast group */
-	mcid = nl_get_multicast_id(state->nl_sock, "nl80211", "regulatory");
-	if (mcid >= 0) {
-		ret = nl_socket_add_membership(state->nl_sock, mcid);
-		if (ret)
-			return ret;
-	}
-
-	/* no sequence checking for multicast messages */
-	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
-
-	if (n_waits && waits) {
-		wait_ev.cmds = waits;
-		wait_ev.n_cmds = n_waits;
-		nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, wait_event, &wait_ev);
-	} else {
-		nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, print_event, NULL);
-	}
-
-	wait_ev.cmd = 0;
-
-	while (!wait_ev.cmd)
-		nl_recvmsgs(state->nl_sock, cb);
-
-	nl_cb_put(cb);
-
-	return wait_ev.cmd;
+	return __handle_cmd(state, idby, argc, argv, NULL);
 }
 
 int main(int argc, char **argv)
 {
 	struct nl80211_state nlstate;
 	int err;
-	const char *argv0;
+	struct cmd *cmd = NULL;
 
 	/* calculate command size including padding */
-	cmd_size = abs((long)&__cmd_NULL_1_CIB_NONE_0
-	             - (long)&__cmd_NULL_0_CIB_NONE_0);
+	cmd_size = abs((long)&__cmd_NULL_NULL_1_CIB_NONE_0
+	             - (long)&__cmd_NULL_NULL_0_CIB_NONE_0);
 	/* strip off self */
 	argc--;
 	argv0 = *argv++;
 
 	if (argc > 0 && strcmp(*argv, "--debug") == 0) {
-		debug = 1;
+		iw_debug = 1;
 		argc--;
 		argv++;
 	}
@@ -502,8 +414,9 @@ int main(int argc, char **argv)
 		return 0;
 	}
 
+	/* need to treat "help" command specially so it works w/o nl80211 */
 	if (argc == 0 || strcmp(*argv, "help") == 0) {
-		usage(argv0);
+		usage(argc != 0);
 		return 0;
 	}
 
@@ -511,30 +424,36 @@ int main(int argc, char **argv)
 	if (err)
 		return 1;
 
-	if (strcmp(*argv, "event") == 0) {
-		if (argc != 1)
-			err = 1;
-		else
-			err = listen_events(&nlstate, 0, NULL);
-	} else if (strcmp(*argv, "dev") == 0) {
+	if (strcmp(*argv, "dev") == 0 && argc > 1) {
 		argc--;
 		argv++;
-		err = handle_cmd(&nlstate, II_NETDEV, argc, argv);
-	} else if (strncmp(*argv, "phy", 3) == 0) {
+		err = __handle_cmd(&nlstate, II_NETDEV, argc, argv, &cmd);
+	} else if (strncmp(*argv, "phy", 3) == 0 && argc > 1) {
 		if (strlen(*argv) == 3) {
 			argc--;
 			argv++;
-			err = handle_cmd(&nlstate, II_PHY_NAME, argc, argv);
+			err = __handle_cmd(&nlstate, II_PHY_NAME, argc, argv, &cmd);
 		} else if (*(*argv + 3) == '#')
-			err = handle_cmd(&nlstate, II_PHY_IDX, argc, argv);
+			err = __handle_cmd(&nlstate, II_PHY_IDX, argc, argv, &cmd);
 		else
-			err = 1;
-	} else
-		err = handle_cmd(&nlstate, II_NONE, argc, argv);
+			goto detect;
+	} else {
+		int idx;
+		enum id_input idby = II_NONE;
+ detect:
+		if ((idx = if_nametoindex(argv[0])) != 0)
+			idby = II_NETDEV;
+		else if ((idx = phy_lookup(argv[0])) >= 0)
+			idby = II_PHY_NAME;
+		err = __handle_cmd(&nlstate, idby, argc, argv, &cmd);
+	}
 
-	if (err == 1)
-		usage(argv0);
-	if (err < 0)
+	if (err == 1) {
+		if (cmd)
+			usage_cmd(cmd);
+		else
+			usage(false);
+	} else if (err < 0)
 		fprintf(stderr, "command failed: %s (%d)\n", strerror(-err), err);
 
 	nl80211_cleanup(&nlstate);
