@@ -1,6 +1,7 @@
 #ifndef __IW_H
 #define __IW_H
 
+#include <stdbool.h>
 #include <netlink/netlink.h>
 #include <netlink/genl/genl.h>
 #include <netlink/genl/family.h>
@@ -84,17 +85,29 @@ extern int iw_debug;
 
 int handle_cmd(struct nl80211_state *state, enum id_input idby,
 	       int argc, char **argv);
+
+struct print_event_args {
+	bool frame, time;
+};
+
 __u32 listen_events(struct nl80211_state *state,
 		    const int n_waits, const __u32 *waits);
+__u32 __listen_events(struct nl80211_state *state,
+		      const int n_waits, const __u32 *waits,
+		      struct print_event_args *args);
 
 
 int mac_addr_a2n(unsigned char *mac_addr, char *arg);
 int mac_addr_n2a(char *mac_addr, unsigned char *arg);
 
+int parse_keys(struct nl_msg *msg, char **argv, int argc);
+
 const char *iftype_name(enum nl80211_iftype iftype);
 int ieee80211_channel_to_frequency(int chan);
 int ieee80211_frequency_to_channel(int freq);
 
+void print_ssid_escaped(const uint8_t len, const uint8_t *data);
+
 int nl_get_multicast_id(struct nl_sock *sock, const char *family, const char *group);
 
 char *reg_initiator_to_string(__u8 initiator);
