@@ -1,7 +1,9 @@
 #!/bin/sh
 
-VERSION="0.9.11"
-OUT="version.h"
+VERSION="0.9.15"
+OUT="$1"
+
+echo '#include "iw.h"' > "$OUT"
 
 if head=`git rev-parse --verify HEAD 2>/dev/null`; then
 	git update-index --refresh --unmerged > /dev/null
@@ -10,13 +12,13 @@ if head=`git rev-parse --verify HEAD 2>/dev/null`; then
 	# on git builds check that the version number above
 	# is correct...
 	[ "${descr%%-*}" = "v$VERSION" ] || exit 2
-	
-	echo -n '#define IW_VERSION "' > "$OUT"
-	echo -n "${descr#v}" >> "$OUT"
+
+	v="${descr#v}"
 	if git diff-index --name-only HEAD | read dummy ; then
-		echo -n "-dirty" >> "$OUT"
+		v="$v"-dirty
 	fi
-	echo '"' >> "$OUT"
 else
-echo "#define IW_VERSION \"$VERSION-nogit\"" > "$OUT"
+	v="$VERSION"
 fi
+
+echo "const char iw_version[] = \"$v\";" >> "$OUT"
